# -*- coding: utf-8 -*-

from __future__ import absolute_import

# real (choose precision here)
ctypedef double RTYPE_t

# complex (choose precision here)
ctypedef double complex ZTYPE_t

# problem data (choose real or complex here)
ctypedef RTYPE_t DTYPE_t

